# -*- coding: utf-8 -*-

__version__ = "0.12.0"
__author__ = "Christopher Goes"
__email__ = "<goes8945@vandals.uidaho.edu>"
__license__ = "Apache 2.0"
__url__ = "https://github.com/GhostofGoes/ADLES"
